/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "BufferPool.h"
#include "vpi_api.h"

namespace xma {

#define XMA_BUFFER_POOL "BufferPool"

//////////////////////////////////////////////////////////////////////////////
BufferPool::BufferPool(VpiMemHandle mem, LogPtr log, uint32_t num, uint32_t size, uint32_t flag, bool auto_extend) : m_log(log), m_auto_extend(auto_extend) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::BufferPool(num = %u, size = %u)", num, size);
  VpiBufferPool* new_pool = nullptr;
  VpiRet         ret      = vpi_buffer_pool_create(mem, num, size, flag, false, 0, &new_pool);
  if (ret < VPI_SUCCESS) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_BUFFER_POOL, "BufferPool::BufferPool() Failed to create buffer pool: %s!", vpi_error_str(ret));
  }
  m_pool = pool_t(new_pool, [](VpiBufferPool* pool) {
    for (uint32_t i = 0; i < pool->num; ++i) {
      VpiBuffer* buf = vpi_buffer_pool_get_list_item(pool, i);
      if (pool->size == 0) {
        buf->m_handle = nullptr;
      }
      if (pool->md_size == 0) {
        buf->m_metadata_handle = nullptr;
      }
      buf->data   = nullptr;
      buf->opaque = nullptr;
      for (int32_t j = 0; j < VPE_CHANNEL_NUM; j++) {
        buf->consume_opaque[j] = nullptr;
      }
    }
    vpi_buffer_pool_unref(pool);
  });
  if (!m_pool) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_BUFFER_POOL, "BufferPool::BufferPool() Out of memory!");
  }

  vpi_buffer_pool_set_capacity(m_pool.get(), MAX_POOL_ITEM_NUM);

  InitData();

  m_pool->opaque = this;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::BufferPool() - returning");
}

///////////////////////////////////////////////////////////////////////////////
BufferPool::~BufferPool() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::~BufferPool()");

  m_pool->opaque = nullptr;
}

///////////////////////////////////////////////////////////////////////////////
int32_t BufferPool::SetLog(LogPtr log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::SetLog()");
  lock_t guard(m_mutex);
  m_log = log;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::SetLog() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
buffer_t BufferPool::GetBuffer() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::GetBuffer()");
  lock_t guard(m_mutex);

  InitData();

  VpiBuffer* buf = nullptr;
  while (1) {
    buf = vpi_buffer_pool_alloc_item(m_pool.get());
    if (!buf) {
      if (m_auto_extend) {
        if (m_pool->num == MAX_POOL_ITEM_NUM) {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_BUFFER_POOL, "BufferPool::GetBuffer() Unable to auto-extend buffer pool!");
        }
        guard.unlock();
        Extend(BUFFER_POOL_AUTO_EXTEND_SIZE);
        guard.lock();
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::GetBuffer() Waiting for free buffer...");
        m_wait_for_empty_buf = 1;
        m_cond.wait(guard);
      }
    } else {
      break;
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::GetBuffer() - returning");
  return buffer_t(buf, [](VpiBuffer* buffer) {
    VpiBufferPool* vpiPool = static_cast<VpiBufferPool*>(buffer->pool);
    BufferPool*    pool    = static_cast<BufferPool*>(vpiPool->opaque);
    if (pool) {
      pool->FreeBuffer(buffer);
    }
  });
}

///////////////////////////////////////////////////////////////////////////////
VpiBuffer* BufferPool::GetBufferAt(uint32_t index) const {
  return vpi_buffer_pool_get_list_item(m_pool.get(), index);
}

///////////////////////////////////////////////////////////////////////////////
VpiBufferPool* BufferPool::GetVpiPool() const {
  return m_pool.get();
}

///////////////////////////////////////////////////////////////////////////////
void BufferPool::FreeBuffer(VpiBuffer* buffer) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::FreeBuffer()");
  lock_t guard(m_mutex);
  if (!vpi_buffer_pool_free_item(m_pool.get(), buffer)) {
    if (m_wait_for_empty_buf == 1) {
      m_cond.notify_one();
      m_wait_for_empty_buf = 0;
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::FreeBuffer() - returning");
}

///////////////////////////////////////////////////////////////////////////////
int32_t BufferPool::Extend(uint32_t num) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::Extend(num = %u)", num);
  lock_t guard(m_mutex);
  if (vpi_buffer_pool_extend(m_pool.get(), num)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_BUFFER_POOL, "BufferPool::Extend() Unable to extend buffer pool!");
  }

  InitData();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_BUFFER_POOL, "BufferPool::Extend() - returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
uint32_t BufferPool::GetCurrentSize() const {
  return m_pool->num;
}

///////////////////////////////////////////////////////////////////////////////
void BufferPool::InitData() {
  for (uint32_t i = m_size; i < m_pool->num; ++i) {
    VpiBuffer* buf = vpi_buffer_pool_get_list_item(m_pool.get(), i);
    if (buf->m_handle) {
      buf->data = reinterpret_cast<void*>(buf->m_handle->bus_address_ep);
    }
    buf->freebuf = nullptr;
  }
  m_size = m_pool->num;
}

} // namespace xma
