/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include "xmacpp/SessionFactory.h"
#include "Exception.h"
#include "XmaLog.h"
#include "BufferPool.h"
#include "vpi_types.h"
#include "vpi_api.h"
#include "vpi_memory.h"
#ifdef __cplusplus
extern "C" {
#endif
#include "vpi_buffer.h"
#ifdef __cplusplus
}
#endif

namespace xma {

class InternalSessionFactory final : public SessionFactory {
public:
  InternalSessionFactory(LogPtr log, XmaInitParameter* param);
  virtual ~InternalSessionFactory();

  // Factory for creating a sessions that contains a plugin
  virtual DecoderSessionPtr   MakeDecoderSession(XmaDecoderProperties* props) override;
  virtual ScalerSessionPtr    MakeScalerSession(XmaScalerProperties* props) override;
  virtual FilterSessionPtr    MakeFilterSession(XmaFilterProperties* props) override;
  virtual MioFilterSessionPtr MakeMioFilterSession(XmaMioFilterProperties* props) override;
  virtual EncoderSessionPtr   MakeEncoderSession(XmaEncoderProperties* props) override;

  virtual SideDataPtr        AllocSideData(enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) override;
  virtual FramePropertiesPtr AllocFrameProperties(enum XmaFormatType format, enum XmaFormatType sw_format, int32_t width, int32_t height, uint32_t flags) override;
  virtual FramePtr           AllocFrame(FramePropertiesPtr props, XmaFraction frame_rate, int64_t pts, bool is_idr) override;
  virtual FramePtr           CloneFrame(
                FramePropertiesPtr props, uint8_t** frame_data, std::function<void(uint8_t**)> free_callback, XmaFraction frame_rate, int64_t pts, bool is_idr) override;
  virtual FramePtr      CloneFrame(FramePtr frame, XmaFraction frame_rate, int64_t pts, bool is_idr) override;
  virtual DataBufferPtr AllocDataBuffer(size_t size, bool dummy, int64_t pts, int64_t poc) override;
  virtual DataBufferPtr CloneDataBuffer(uint8_t* data, size_t size, std::function<void(uint8_t*)> free_callback, int64_t pts, int64_t poc) override;

  LogPtr       GetLog();
  VpiDevHandle GetDevHandle() const;
  VpiMemHandle GetMemHandle() const;
  VpiMemHandle CreateMemHandle() const;
  buffer_t     GetBuffer(uint32_t size, XmaBufferType type);

#ifdef XMA_CHECK_MEM
  std::mutex                   m_mutex;
  std::list<XmaFrame*>         m_frames;
  std::list<XmaDataBuffer*>    m_data_buffers;
  std::list<XmaFrameSideData*> m_side_data;
#endif

protected:
  using vpi_handle_t      = std::unique_ptr<void, std::function<void(VpiHandle ctx)>>;
  using vpi_dev_handle_t  = std::unique_ptr<void, std::function<void(VpiDevHandle dev_ctx)>>;
  using vpi_mem_handle_t  = std::unique_ptr<void, std::function<void(VpiMemHandle mem_ctx)>>;
  using buffer_pool_map_t = std::map<uint32_t, BufferPoolPtr>;
  LogPtr            m_log;
  vpi_handle_t      m_ctx;
  vpi_dev_handle_t  m_dev_ctx;
  vpi_mem_handle_t  m_mem_ctx;
  char              m_device[64] = {'\0'};
  buffer_pool_map_t m_buffers[3];
};

using IntSessionFactoryPtr = std::shared_ptr<InternalSessionFactory>;

}; // Namespace xma
